package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CityGmlUtils;
import de.hft.stuttgart.citydoctor2.utils.CopyHandler;
import de.hft.stuttgart.citydoctor2.utils.Copyable;
import org.citygml4j.core.util.geometry.GeometryFactory;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurface;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurfaceProperty;
import org.xmlobjects.gml.model.geometry.primitives.Solid;
import org.xmlobjects.gml.model.geometry.primitives.SolidProperty;

import java.io.Serial;
import java.util.ArrayList;
import java.util.List;

/**
 * Represents all types of furniture used inside Buildings.
 */
public abstract class AbstractFurniture extends CityObject {

    @Serial
    private static final long serialVersionUID = -9050689238027190674L;

    private final List<BoundarySurface> boundarySurfaceList = new ArrayList<>();

    private CityObject parent;

    private org.citygml4j.core.model.construction.AbstractFurniture af;

    @Override
    public void accept(Check c) {
        super.accept(c);
        if (c.canExecute(this)) {
            c.check(this);
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.accept(c);
        }
    }

    @Override
    public void collectContainedErrors(List<CheckError> errors) {
        super.collectContainedErrors(errors);
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.collectContainedErrors(errors);
        }
    }

    @Override
    public void clearAllContainedCheckResults() {
        super.clearAllContainedCheckResults();
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.clearAllContainedCheckResults();
        }
    }

    @Override
    public boolean containsError(CheckId checkIdentifier) {
        boolean hasError = super.containsError(checkIdentifier);
        if (hasError) {
            return true;
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            if (boundarySurface.containsError(checkIdentifier)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public boolean containsAnyError() {
        boolean hasError = super.containsAnyError();
        if (hasError) {
            return true;
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            if (boundarySurface.containsAnyError()) {
                return true;
            }
        }
        return false;
    }


    @Override
    public org.citygml4j.core.model.construction.AbstractFurniture getGmlObject() {
        return af;
    }

    public void addBoundarySurface(BoundarySurface boundarySurface) {
        boundarySurfaceList.add(boundarySurface);
        boundarySurface.setParent(this);
    }

    public List<BoundarySurface> getBoundarySurfaceList() {
        return boundarySurfaceList;
    }

    @Override
    public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
        for (Geometry geom : getGeometries()) {
            if (geom instanceof ImplicitGeometryHolder) {
                continue;
            }
            if (geom.getType() == GeometryType.MULTI_SURFACE) {
                MultiSurface ms = CityGmlUtils.createMultiSurface(geom, factory, config);
                setMultiSurfaceAccordingToLod(geom, ms);
            } else {
                Solid solid = CityGmlUtils.createSolid(geom, factory, config);
                setSolidAccordingToLod(geom, solid);
            }
        }
    }

    protected void setGmlObject(org.citygml4j.core.model.construction.AbstractFurniture af) {
        this.af = af;
    }

    public void setParent(CityObject co) {
        parent = co;
    }


    @Override
    public void unsetGmlGeometries() {
        af.setLod0MultiSurface(null);
        af.setLod2MultiSurface(null);
        af.setLod3MultiSurface(null);
        af.setLod1Solid(null);
        af.setLod2Solid(null);
        af.setLod3Solid(null);
    }

    private void setMultiSurfaceAccordingToLod(Geometry geom, MultiSurface ms) {
        switch (geom.getLod()) {
            case LOD0:
                af.setLod0MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD2:
                af.setLod2MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD3:
                af.setLod3MultiSurface(new MultiSurfaceProperty(ms));
                break;
            default:
                throw new IllegalStateException("Cannot add " + geom.getLod() + " multi surface to buildings");
        }
    }

    private void setSolidAccordingToLod(Geometry geom, Solid solid) {
        switch (geom.getLod()) {
            case LOD1:
                af.setLod1Solid(new SolidProperty(solid));
                break;
            case LOD2:
                af.setLod2Solid(new SolidProperty(solid));
                break;
            case LOD3:
                af.setLod3Solid(new SolidProperty(solid));
                break;
            default:
                throw new IllegalStateException("Cannot add " + geom.getLod() + " solid to buildings");
        }
    }


    @Override
    public void prepareForChecking() {
        super.prepareForChecking();
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.prepareForChecking();
        }
    }

    @Override
    public void clearMetaInformation() {
        super.clearMetaInformation();
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.clearMetaInformation();
        }
    }


    @Override
    public FeatureType getFeatureType() {
        return FeatureType.FURNITURE;
    }

    @Override
    public void fillValues(Copyable original, CopyHandler handler) {
        super.fillValues(original, handler);
        AbstractFurniture originalAf = (AbstractFurniture) original;
        af = originalAf.af;
    }
}
