package de.hft.stuttgart.citydoctor2.healing;

import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.hft.stuttgart.citydoctor2.check.HealingID;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.check.error.MultipleConnectedComponentsError;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.math.Plane;
import de.hft.stuttgart.citydoctor2.math.Vector3d;

public class HealWrongDormers implements HealingMethod {

	private static final Logger logger = LogManager.getLogger(HealWrongDormers.class);

	private static final double EPSILON = 0.2;

	@Override
	public boolean visit(MultipleConnectedComponentsError err, ModificationListener l) {
		logger.debug("Executing Repair for MultipleConnectedComponentsError");
		for (int i = 0; i < err.getComponents().size() - 1; i++) {
			List<Polygon> component1 = err.getComponents().get(i);
			for (int j = i + 1; j < err.getComponents().size(); j++) {
				List<Polygon> component2 = err.getComponents().get(j);
				if (checkForOverlappingPolygons(component1, component2)) {
					return true;
				}
			}
		}
		return false;
	}

	private boolean checkForOverlappingPolygons(List<Polygon> component1,
			List<Polygon> component2) {
		for (Polygon p1 : component1) {
			for (Polygon p2 : component2) {
				if (isPolygonInOtherPolygon(p1, p2)) {
					// p1 is in p2
					createInnerRing(p1, p2);
					return true;
				}
				if (isPolygonInOtherPolygon(p2, p1)) {
					// p2 is in p1
					createInnerRing(p2, p1);
					return true;
				}
			}
		}
		return false;
	}

	private void createInnerRing(Polygon poly1, Polygon poly2) {
		LinearRing innerRing = new LinearRing(LinearRingType.INTERIOR);
		poly2.addInteriorRing(innerRing);
		for (Vertex v : poly1.getExteriorRing().getVertices()) {
			innerRing.addVertex(v);
		}
		poly1.remove();
	}
	
	private boolean isPolygonInOtherPolygon(Polygon poly1, Polygon poly2) {
		Plane plane = calculatePlane(poly2);
		for (Vertex v : poly1.getExteriorRing().getVertices()) {
			if (plane.getDistance(v) > EPSILON || !poly2.isPointInsideExteriorRing(v)) {
				return false;
			}
		}
		return true;
	}
	
	private Plane calculatePlane(Polygon p) {
		Vector3d normal = p.calculateNormalNormalized();
		return new Plane(normal, p.getExteriorRing().getVertices().get(0));
	}

	@Override
	public HealWrongDormers createNew() {
		return new HealWrongDormers();
	}

	@Override
	public HealingID getID() {
		return HealingID.S_WRONG_DORMERS;
	}

}
